import glob
import importlib
import json
import os
import re
import warnings
from functools import partial
from typing import Callable, Dict, List, Literal

import numpy as np
import requests
import xarray as xr
from osgeo import gdal

import pywapor
from pywapor.collect.downloader import collect_sources
from pywapor.enhancers.temperature import lapse_rate_to_all
from pywapor.general.compositer import time_bins
from pywapor.general.logger import adjust_logger, log
from pywapor.general.processing_functions import (
    adjust_timelim_dtype,
    func_from_string,
    has_wrong_bb_or_period,
    is_corrupt_or_empty,
    open_ds,
    remove_ds,
)


class Configuration:
    """A `pywapor.configuration` specifies which products to use for each variable and
    how to preprocess them."""

    variable_categories = {
        "optical": [
            "ndvi",
            "r0",
            "mndwi",
            "psri",
            "vari_red_edge",
            "bsi",
            "nmdi",
            "green",
            "nir",
        ],
        "thermal": ["bt", "lst"],
        "solar radiation": ["ra_flat"],
        "precipitation": ["p"],
        "elevation": ["z", "slope", "aspect"],
        "meteorological": [
            "t_air",
            "t_air_min",
            "t_air_max",
            "u",
            "vp",
            "p_air",
            "p_air_0",
            "u2m",
            "v2m",
            "qv",
            "wv",
            "t_dew",
        ],
        "statics": [
            "lw_slope",
            "lw_offset",
            "z_obst_max",
            "rs_min",
            "land_mask",
            "vpd_slope",
            "t_opt",
            "t_amp",
            "t_amp_year",
            "rn_slope",
            "rn_offset",
            "z_oro",
        ],
        "soil moisture": ["se_root"],
    }
    """Specifies which variables area included in each summary group."""

    category_variables = {}
    """Specifies a category per variable, i.e. its the inverse of `variable_categories` and
    any changes made should be reflected in both."""
    for x in [
        {var: cat for var in varis} for cat, varis in variable_categories.items()
    ]:
        category_variables = {**category_variables, **x}

    se_root_vars = [
        [("ndvi",)],
        [("bt", "lst"), ("lst",), ("bt",)],
        [("u",), ("u2m", "v2m")],
        [("p_air",)],
        [("p_air_0",)],
        [("t_air",)],
        [("wv",)],
        [("t_dew",), ("qv",)],
    ]
    """Specifies the required variables to run the full SE_ROOT model and gives
    different options. For example, it possible to either use `"t_dew"` or `"qv"`.
    Another example, it is possible to either use `"bt"` AND `"lst"` or only 
    one of them."""

    et_look_vars = [
        [("ndvi",)],
        [("r0",)],
        [("se_root",)],
        [("p",)],
        [("z", "slope", "aspect"), ("z")],
        [("ra_flat",)],
        [("u",), ("u2m", "v2m")],
        [("p_air",)],
        [("p_air_0",)],
        [("t_air",)],
        [("t_air_min",)],
        [("t_air_max",)],
        [("t_dew",), ("qv",)],
        [("rn_offset",)],
        [("rn_slope",)],
        [
            ("t_amp",),
        ],
        [("t_opt",)],
        [("vpd_slope",)],
        [("lw_offset",)],
        [("lw_slope",)],
        [("rs_min",), ()],
        [("land_mask",), ()],
        [("z_obst_max",), ()],
        [("z_oro",), ()],
    ]
    """Specifies the required variables to run the full ET_LOOK model and gives
    different options. For example, it possible to either use `"t_dew"` or `"qv"`.
    Another example, it is possible to either use `"u"` or `"u2m"` AND `"v2m"`."""

    @staticmethod
    def source_func(x: str):
        """Given a full `source.product` string, returns the `source` only.

        Parameters
        ----------
        x : str
            String describing the `source.product`.

        Returns
        -------
        str
            The `source`.
        """
        return x if "FILE:" in x else x.split(".")[0]

    @staticmethod
    def pname_func(x: str):
        """Given a full `source.product` string, returns the `product` only.

        Parameters
        ----------
        x : str
            String describing the `source.product`.

        Returns
        -------
        str
            The `product`.
        """
        return "none" if "FILE:" in x else ".".join(x.split(".")[1:])

    def __init__(
        self,
        full: dict | None = None,
        summary: dict | None = None,
        se_root: dict | None = None,
        et_look: dict | None = None,
    ):
        """It is advised to use one of the `from_` methods to instantiate a configuration
        instead.

        Parameters
        ----------
        full : dict | None, optional
            The full configuration contains all variables that will be used by either
            ET_LOOK or SE_ROOT. When changed, `configuration.update_se_root_config` and
            `configuration.update_et_look_config` should be called to make sure they are
            in sync with eachother, by default None.
        summary : dict | None, optional
            A summary (and a simplification) of the `full` configuration, by default None.
        se_root : dict | None, optional
            Part of the `full` configuration that is relevant for the SE_ROOT model. Should
            generally not be changed directly, but through changing `configuration.full` and calling
            `configuration.update_se_root_config` instead, by default None.
        et_look : dict | None, optional
            Part of the `full` configuration that is relevant for the ET_LOOK model. Should
            generally not be changed directly, but through changing `configuration.full` and calling
            `configuration.update_et_look_config` instead, by default None.

        Notes
        -------
        See `configuration.full` for more details and the `configuration.from_summary` method
        for a description of `configuration.summary`.
        """

        self.full = full
        """The full configuration contains all variables that will be used by either
        ET_LOOK or SE_ROOT. When changed, `configuration.update_se_root_config` and
        `configuration.update_et_look_config` should be called to make sure they are
        in sync with eachother.
        
        Example
        -------
        Each `key` in `full` defines (1) the `products` from which the variable should
        be generated, (2) the `temporal_interp` that should be used, (3) any specific enhancers
        that should be applied to this variable, (4) which spatial interpolation should be applied
        and (5) how the variable should be composited.::

            full["ndvi"] = {
                'products': [
                    {
                    'source': 'SENTINEL2',
                    'product_name': 'S2MSI2A_R60m',
                    'enhancers': [
                        {'func': 'pywapor.collect.product.SENTINEL2.calc_normalized_difference'},
                        {'func': 'pywapor.enhancers.gap_fill.gap_fill'}
                        ],
                    'is_example': True
                    },
                ],
                'temporal_interp': {'lmbdas': 1000.0, 'method': 'whittaker'},
                'variable_enhancers': [],
                'spatial_interp': 'bilinear',
                'composite_type': 'mean'
            }
        """

        self.summary = summary
        """A summary (and a simplification) of the `full` configuration."""

        self.se_root = se_root
        """Part of the `full` configuration that is relevant for the SE_ROOT model. Should 
        generally not be changed directly, but through changing `configuration.full` and calling
        `configuration.update_se_root_config` instead."""

        self.et_look = et_look
        """Part of the `full` configuration that is relevant for the ET_LOOK model. Should 
        generally not be changed directly, but through changing `configuration.full` and calling
        `configuration.update_et_look_config` instead."""

    def __repr__(self):
        summary = self.summary.copy()

        example_product = summary.pop("_EXAMPLE_")
        whittaker = summary.pop("_WHITTAKER_")
        sharpen = summary.pop("_ENHANCE_")

        base = "\n".join(
            [
                f"--> {cat.title()} data:\n    > `{'`, `'.join(sources)}`"
                for cat, sources in summary.items()
            ]
        )

        additional = f"""
--> Example Product:
    > {example_product}
--> Applying whittaker to variables from sources:
    > `{"` and `".join(whittaker)}`
--> Sharpening these variables:
    > `{"` and `".join(sharpen)}`"""
        return base + additional

    @staticmethod
    def default_enhancers(src_prod: str, var: str):
        """Given a `source.product` string and a variable name, returns
        which enhancers should be applied by default.

        Parameters
        ----------
        src_prod : str
            `source.product` string, e.g. `"SENTINEL2.S2MSI2A_R20m"`.
        var : str
            variable name, e.g. `"ndvi"`.

        Returns
        -------
        List[Callable]
            List of functions that are to be applied by default to the variable
            after downloading.
        """
        source = Configuration.source_func(src_prod)
        if "FILE:" in source:
            return []
        product_name = Configuration.pname_func(src_prod)
        mod = importlib.import_module(f"pywapor.collect.product.{source}")
        x = mod.default_post_processors(product_name, [var])[var]

        funcs = list()
        for x_ in x:
            if isinstance(x_, partial):
                funcs.append(
                    {
                        "func": f"{x_.func.__module__}.{x_.func.__name__}",
                        "args": x_.args,
                        "keywords": x_.keywords,
                    }
                )
            else:
                funcs.append({"func": f"{x_.__module__}.{x_.__name__}"})

        return funcs

    @classmethod
    def from_name(
        cls,
        name: Literal[
            "WaPOR3_level_2",
            "WaPOR3_level_3",
            "WaPOR2_level_1",
            "WaPOR2_level_2",
            "WaPOR2_level_3",
            "nrt",
            "all_in",
        ],
    ):
        """Create a configuration instance from a `name`.

        Parameters
        ----------
        name : Literal["WaPOR3_level_2", "WaPOR3_level_3", "WaPOR2_level_1", "WaPOR2_level_2", "WaPOR2_level_3", "nrt", "all_in"]
            Choose which configuration to instantiate.

        Returns
        -------
        Configuration
            A pywapor.Configuration.

        Raises
        ------
        ValueError
            Unkown `name`.
        """
        log.info(f"--> Searching configuration for `{name}`.").add()

        folder = os.path.realpath(pywapor.__path__[0])
        available_configs = glob.glob(os.path.join(folder, "configs", "*.json"))
        available_names = sorted(
            [os.path.splitext(os.path.split(x)[-1])[0] for x in available_configs]
        )
        log.info(f"> Available configurations are `{'`, `'.join(available_names)}`")

        synonyms = {
            "level_1": "WaPOR2_level_1",
            "level_2": "WaPOR2_level_2",
            "level_3": "WaPOR2_level_3",
            "level_1_v2": "WaPOR2_level_1",
            "level_2_v2": "WaPOR2_level_2",
            "level_3_v2": "WaPOR2_level_3",
            "level_1_v3": "WaPOR3_level_1",
            "level_2_v3": "WaPOR3_level_2",
            "level_3_v3": "WaPOR3_level_3",
        }

        name_ = synonyms.get(name, name)

        selection = None
        for fh in available_configs:
            fn = os.path.splitext(os.path.split(fh)[-1])[0]
            if fn == name_:
                selection = fh
                break

        if isinstance(selection, type(None)):
            raise ValueError(
                f"No configuration found for name `{name}`, choose one from `{'`, `'.join(available_names)}`."
            )
        else:
            config = Configuration.from_json(selection)

        log.sub().info("--> Configuration set.")

        return config

    def to_json(self, fh: str):
        """Save the configuration as a json file.

        Parameters
        ----------
        fh : str
            Path to destination.

        """
        out = {
            "summary": self.summary,
            "full": self.full,
            "se_root": self.se_root,
            "et_look": self.et_look,
        }

        class EncodeSetTuple(json.JSONEncoder):
            def encode(self, obj):
                def hint_tuples(item):
                    if isinstance(item, tuple):
                        return {"type": "tuple", "value": list(item)}
                    if isinstance(item, list):
                        return [hint_tuples(e) for e in item]
                    if isinstance(item, set):
                        return {"type": "set", "value": list(item)}
                    if isinstance(item, dict):
                        return {key: hint_tuples(value) for key, value in item.items()}
                    else:
                        return item

                return super(EncodeSetTuple, self).encode(hint_tuples(obj))

        json_string = json.dumps(out, cls=EncodeSetTuple)
        with open(fh, "w", encoding="utf8") as x:
            x.write(json_string)

    @classmethod
    def from_json(cls, fh: str):
        """Create a configuration instance from a `json`-file.

        Parameters
        ----------
        fh : str
            Path of the JSON-file.

        Returns
        -------
        Configuration
            A pywapor.Configuration.
        """
        log.info(f"--> Loading configuration from `{os.path.split(fh)[1]}`.").add()

        def decode_set(dct):
            if dct.get("type", None) == "set":
                return set(dct.get("value"))
            if dct.get("type", None) == "tuple":
                return tuple(dct.get("value"))
            return dct

        with open(fh, "r", encoding="utf8") as fp:
            config_ = json.load(fp, object_hook=decode_set)

        config = cls(
            full=config_["full"],
            summary=config_["summary"],
            se_root=config_["se_root"],
            et_look=config_["et_look"],
        )

        config.validate()
        log.sub().info("--> Configuration loaded.")
        return config

    @classmethod
    def from_summary(cls, summary: dict):
        """Create a configuration instance from a summary.

        Parameters
        ----------
        summary : dict
            A summary of a configuration, see after this for a description
            of the different keys andthe example section below.
        \\_EXAMPLE_ : str
            Specifies which product should be use for spatial alignment. The
            product defined here will determine the resolution
            of the final output.
        \\_ENHANCE_ : Dict[str, list]
            Specifies which functions should be applied to specific variables,
            is generally used to turn on or off thermal sharpening.
        \\_WHITTAKER_ : Dict[str, Dict]
            Specifies which products should be interpolated with the special
            whittaker interpolation. I.e the interpolation will be applied to
            all variables belong to the specified product.
        elevation, meteorological, statics, etc. : set
            Specify which products to use for the category of variables. Check
            `pywapor.Configuration.variable_categories` to see (or adjust) which
            variables belong to each category.

        Returns
        -------
        Configuration
            A pywapor.Configuration.

        Example
        -------
        The summary can contain several keys, of which only `_EXAMPLE_`,
        `_ENHANCE_` and `_WHITTAKER_` are mandatory. Additionaly, the product
        defined as `_EXAMPLE_` should also appear in at least one of the other
        keys (e.g. below the example product `'SENTINEL2.S2MSI2A_R20m'` also
        appears under the `optical` key).::

            summary = {
                'elevation': {'COPERNICUS.GLO30'},
                'meteorological': {'GEOS5.tavg1_2d_slv_Nx'},
                'optical': {'SENTINEL2.S2MSI2A_R20m'},
                'precipitation': {'CHIRPS.P05'},
                'solar radiation': {'ERA5.sis-agrometeorological-indicators'},
                'statics': {'STATICS.WaPOR3'},
                'thermal': {'VIIRSL1.VNP02IMG'},

                'soil moisture': {'FILE:{folder}{sep}se_root_out*.nc'},

                '_EXAMPLE_': 'SENTINEL2.S2MSI2A_R20m',
                '_ENHANCE_': {"bt": ["pywapor.enhancers.dms.thermal_sharpener.sharpen"],},
                '_WHITTAKER_': {
                    'SENTINEL2.S2MSI2A_R20m': {'lmbdas': 1000.0, 'method': 'whittaker'},
                    'VIIRSL1.VNP02IMG': {'a': 0.85, 'lmbdas': 1000.0, 'method': 'whittaker'}},
                }

        Finally, instead of specifying a `source.product` string, it is also possible to
        pass in your own files, as long as they contain the required variables. To do so, create
        a string that can be passed to `glob.glob` and append `"FILE:"` in front of it. See the
        `soil_moisture` key above for an example.
        """
        log.info("--> Creating configuration from summary.").add()

        valids = set(cls.variable_categories.keys()).union(
            {"_EXAMPLE_", "_WHITTAKER_", "_ENHANCE_"}
        )
        invalid_keys = set(summary.keys()).difference(valids)
        if invalid_keys:
            log.warning(
                f"--> Summary contains unrecognized keys ({invalid_keys}), they will be ignored."
            )
            summary = {k: v for k, v in summary.items() if k not in invalid_keys}

        example_product = summary.pop("_EXAMPLE_", "")
        temporal_interp = summary.pop("_WHITTAKER_", {})
        enhance = summary.pop("_ENHANCE_", {})

        full = dict()

        unique_enhancers = set([item for row in list(enhance.values()) for item in row])
        extra_vars = list()
        for enhancer in unique_enhancers:
            func = func_from_string(enhancer)
            extra_vars += [[(x,)] for x in func.__kwdefaults__["req_vars"]]

        for var_group in cls.et_look_vars + cls.se_root_vars + extra_vars:
            for var_ in var_group:
                for var in var_:
                    if var in full.keys():
                        continue

                    cat = cls.category_variables[var]
                    possible_prods = summary.get(cat, None)

                    if isinstance(possible_prods, type(None)):
                        continue

                    products = []
                    t_interps = list()
                    for prod in possible_prods:
                        valid = cls.has_var(prod, var)
                        if valid:
                            enhancers = cls.default_enhancers(prod, var)

                            products += [
                                {
                                    "source": cls.source_func(prod),
                                    "product_name": cls.pname_func(prod),
                                    "enhancers": enhancers,
                                    "is_example": prod == example_product,
                                }
                            ]

                        x = temporal_interp.get(prod, None)
                        if not isinstance(x, type(None)):
                            t_interps += [x]

                    if len(products) == 0:
                        continue

                    t_interps_unique = list(
                        map(
                            json.loads,
                            set(
                                map(lambda x: json.dumps(x, sort_keys=True), t_interps)
                            ),
                        )
                    )
                    # If no particular t_interp has been specified for this variable, use "linear".
                    if len(t_interps_unique) == 0:
                        temporal_interp_ = "linear"
                    # If a single unique t_interp has been specified, use that.
                    elif len(t_interps_unique) == 1:
                        temporal_interp_ = t_interps_unique[0]
                    # If multiple t_interp are specified for one variable, warn and use the first one.
                    else:
                        warn_string = str(
                            {
                                "_WHITTAKER_": {
                                    x: temporal_interp[x] for x in possible_prods
                                }
                            }
                        )
                        log.warning(
                            f"--> Multiple temporal interpolations specified for {var} (through `{warn_string}`), will continue using {t_interps_unique[0]}."
                        )
                        temporal_interp_ = t_interps_unique[0]

                    variable_enhancers = enhance.get(var, [])

                    composite_type = (
                        "mean"
                        if var.split("_")[-1] not in ["min", "max"]
                        else var.split("_")[-1]
                    )

                    full[var] = {
                        "products": products,
                        "temporal_interp": temporal_interp_,
                        "variable_enhancers": variable_enhancers,
                        "spatial_interp": {"aspect": "nearest"}.get(var, "bilinear"),
                        "composite_type": composite_type,
                    }

        config = cls(full=full)
        config.validate()
        config.summarize()
        config.update_se_root_config()
        config.update_et_look_config()
        log.sub().info("--> Configuration created.")
        return config

    @staticmethod
    def has_var(src_prod: str, variable: str, verbose=True):
        """Check if the given `variable` exists for the given `source.product`.

        Parameters
        ----------
        src_prod : str
            `source.product` string, e.g. `"SENTINEL2.S2MSI2A_R20m"`.
        variable : str
            variable name, e.g. `"ndvi"`.
        verbose : bool, optional
            Turn on or off info logging, by default True.

        Returns
        -------
        bool
            Whether or not the `variable` exists.
        """

        source = Configuration.source_func(src_prod)
        product_name = Configuration.pname_func(src_prod)
        if "FILE:" in source:
            if not verbose:
                log.info(
                    f"> Variable `{variable}` will be loaded from a file `{source}`."
                )
            return True
        mod = importlib.import_module(f"pywapor.collect.product.{source}")
        try:
            mod.default_vars(product_name, [variable])
            valid = True
        except TypeError:
            valid = False
            if not verbose:
                log.warning(
                    f"> {source}.{product_name} does not have a variable called `{variable}`."
                )
        return valid

    def validate(self):
        """Checks for each product specified in `configuration.full` if the
        required variable exists.

        Returns
        -------
        bool
            Whether or not the variable sources are valid.
        """
        log.info("--> Validating configuration.").add()
        valids = []
        for var in self.full:
            for product in self.full[var]["products"]:
                prod = f"{product['source']}.{product['product_name']}"
                valid = Configuration.has_var(prod, var, verbose=False)
                valids.append(valid)
        if all(valids):
            log.info("> All specified variables sources are valid.")
        log.sub()
        return valid

    def update_se_root_config(self):
        """Sync `configuration.full` with `configuration.se_root`."""
        log.info("--> Making configuration for SE_ROOT.").add()

        sharpened_vars = {
            var
            for var, config in self.full.items()
            if any(["sharpen" in x for x in config["variable_enhancers"]])
        }

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]

        se_root_config = dict()
        for var_group in self.se_root_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        se_root_config[var_] = self.full[var_].copy()
                        _ = se_root_config[var_].pop("composite_type")
                    break
                elif var == var_group[-1]:
                    substring = [f"[`{'` and `'.join(x)}`]" for x in var_group]
                    log.warning(
                        f"--> No configuration found for {' or '.join(substring)}."
                    )
                else:
                    continue

        to_sharpen = set(sharpened_vars) & set(se_root_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    se_root_config[var] = self.full[var]
                else:
                    log.warning(
                        f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`)."
                    )

        self.se_root = se_root_config
        log.sub()

    def update_et_look_config(self):
        """Sync `configuration.full` with `configuration.et_look`."""
        log.info("--> Making configuration for ET_LOOK.").add()

        sharpened_vars = {
            var
            for var, config in self.full.items()
            if any(["sharpen" in x for x in config["variable_enhancers"]])
        }

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]

        et_look_config = dict()
        for var_group in self.et_look_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        et_look_config[var_] = self.full[var_].copy()
                    if len(var) == 0 and var == var_group[-1]:
                        substring = [f"[`{'` and `'.join(x)}`]" for x in var_group[:-1]]
                        log.info(
                            f"--> No configuration found for {' or '.join(substring)}, will use constant value."
                        )
                    break
                elif var == var_group[-1]:
                    substring = [f"[`{'` and `'.join(x)}`]" for x in var_group]
                    log.warning(
                        f"--> No configuration found for {' or '.join(substring)}."
                    )
                else:
                    continue

        to_sharpen = set(sharpened_vars) & set(et_look_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    et_look_config[var] = self.full[var]
                else:
                    log.warning(
                        f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`)."
                    )

        self.et_look = et_look_config
        log.sub()

    def summarize(self):
        """Creates a summary of `configuration.full`. Note that a `summary` cannot
        show all details of a full configuration and check the `configuration.from_summary`
        method for a more detailed description of a `summary`.
        """

        log.info("--> Making summary of configuration.")

        summary = dict()
        for var, config in self.full.items():
            sources = set()
            for x in config["products"]:
                if "FILE:" in x["source"]:
                    sources.add(x["source"])
                else:
                    sources.add(f"{x['source']}.{x['product_name']}")
            cat = self.category_variables.get(var, "unknown")
            if cat == "other":
                log.warning(f"--> Unknown variable `{var}` found.")
            if cat in list(summary.keys()):
                summary[cat] = summary[cat].union(sources)
            else:
                summary[cat] = sources

        whittaker = dict()
        enhance = dict()
        example = None
        for k, v in self.full.items():
            if isinstance(example, type(None)):
                example_ = [
                    f"{prod['source']}.{prod['product_name']}"
                    for prod in v["products"]
                    if prod["is_example"]
                ]
                if len(example_) > 0:
                    example = example_[0]
            if isinstance(v["temporal_interp"], dict):
                if v["temporal_interp"].get("method", "") == "whittaker":
                    for prod in v["products"]:
                        whittaker[f"{prod['source']}.{prod['product_name']}"] = v[
                            "temporal_interp"
                        ]
            if len(v["variable_enhancers"]) > 0:
                enhance[k] = v["variable_enhancers"]

        summary["_WHITTAKER_"] = whittaker
        summary["_ENHANCE_"] = enhance
        summary["_EXAMPLE_"] = example

        self.summary = summary


class Project:
    """A `pywapor.project` contains all (meta)data required to run one of the included
    models. It is closely linked to a `project_folder` on your disk and different
    projects should never share the same folder.
    """

    def __init__(
        self,
        project_folder: str,
        bb: List[float],
        period: List[str],
        configuration: Configuration = None,
    ):
        assert bb[::2][0] < bb[::2][1], "Invalid Bounding-Box"
        self.lonlim: List[float] = bb[::2]
        """Longitude limits, e.g. [-170.2, -160.78]."""

        assert bb[1::2][0] < bb[1::2][1], "Invalid Bounding-Box"
        self.latlim: List[float] = bb[1::2]
        """Latitude limits, e.g. [-12.9 -79.6]."""

        if not os.path.isdir(project_folder):
            os.makedirs(project_folder)
        self.folder: str = project_folder
        """Folder in which (temporary) files will be stored."""

        self.period: List[str] = adjust_timelim_dtype(period.copy())
        """Period for which data will be generated, e.g. ['2021-03-01', '2021-05-07']."""

        self.configuration: Configuration = configuration
        """Configuration instance describing which sources to use, see pywapor.Configuration."""

        self.dss: dict | None = None
        """Overview of collected datasets."""

        self.bb: List[float] = bb
        """Area for which data will be generated, e.g. [31.0, 28.9, 31.2, 29.1]."""

        os.environ["pyWaPOR_bb"] = str(self.bb)
        os.environ["pyWaPOR_period"] = str(self.period)

        warnings.filterwarnings("ignore", message="invalid value encountered in power")
        adjust_logger(True, self.folder, "INFO")

        self.se_root_in: xr.Dataset | str | None = None
        """Dataset with input for the SE_ROOT model."""
        if os.path.isfile(os.path.join(self.folder, "se_root_in.nc")):
            self.se_root_in = open_ds(os.path.join(self.folder, "se_root_in.nc"))

        self.se_root_out: xr.Dataset | str | None = None
        """Dataset with output of the SE_ROOT model."""
        se_root_outs = glob.glob(os.path.join(self.folder, "se_root_out*.nc"))
        if se_root_outs:
            self.se_root_out = open_ds(max(se_root_outs, key=os.path.getmtime))

        self.et_look_in: xr.Dataset | str | None = None
        """Dataset with input for the ET_LOOK model."""
        if os.path.isfile(os.path.join(self.folder, "et_look_in.nc")):
            self.et_look_in = open_ds(os.path.join(self.folder, "et_look_in.nc"))

        self.et_look_out: xr.Dataset | str | None = None
        """Dataset with output of the ET_LOOK model."""
        et_look_outs = glob.glob(os.path.join(self.folder, "et_look_out*.nc"))
        if et_look_outs:
            self.et_look_out = open_ds(max(et_look_outs, key=os.path.getmtime))

        log.info("> PROJECT").add()
        log.info(self.__repr__())
        self.check_pywapor_version()
        self.check_gdal_drivers()
        log.sub().info("< PROJECT")

    def check_gdal_drivers(self):
        """Check if required GDAL drivers are installed."""
        required_drivers = [
            "GTiff",
            "JP2OpenJPEG",
            "NETCDF",
            "HDF5",
        ]
        log.info(f"--> GDAL ({gdal.__version__}):").add()
        drivers_check = True
        for driver in required_drivers:
            if isinstance(gdal.GetDriverByName(driver), type(None)):
                log.warning(f"> Driver `{driver}` not found.")
                drivers_check = False

        if drivers_check:
            log.info("> All required GDAL drivers found.")
        log.sub()

    def check_pywapor_version(self):
        """Check if current pywapor version is the most recent."""
        current_version = pywapor.__version__
        package = "pywapor"
        log.info(f"--> pyWaPOR ({current_version}):").add()
        try:
            response = requests.get(f"https://pypi.org/pypi/{package}/json")
            response.raise_for_status()
        except Exception as _:
            log.warning("> Unable to check for pyWaPOR updates.")
        else:
            latest_version = response.json()["info"]["version"]
            if latest_version == current_version:
                log.info("> Up to date.")
            else:
                log.warning(f"> Latest version is '{latest_version}'.")
                log.warning("> Please update pywapor.")
        log.sub()

    @staticmethod
    def set_remove_temp_files(remove: bool = True):
        """Set whether or not to `remove` temporary files created during the different
        processes.

        Parameters
        ----------
        remove : bool, optional
            Set to False to NOT remove temporary files, by default True.
        """
        if remove:
            os.environ["PYWAPOR_REMOVE_TEMP_FILES"] = "YES"
        else:
            os.environ["PYWAPOR_REMOVE_TEMP_FILES"] = "NO"

        rmve = {"NO": False, "YES": True}.get(
            os.environ.get("PYWAPOR_REMOVE_TEMP_FILES", "YES"), True
        )
        if rmve:
            log.info("--> Going forward, temporary files will be removed.")
        else:
            log.info("--> Going forward, temporary fiels will NOT be removed.")

    def __repr__(self):
        project_str = f"""--> Project Folder:
        > {self.folder}
    --> Period:
        > {self.period[0]} - {self.period[1]}
    --> Bounding-Box:\n
                 {self.latlim[1]:8.4f}
                 ┌─────────┐
                 │         │
       {self.lonlim[0]:9.4f} │         │{self.lonlim[1]:9.4f}
                 │         │
                 └─────────┘
                 {self.latlim[0]:8.4f}\n
    --> Configuration:
        > {self.configuration}"""
        return project_str

    def load_configuration(
        self,
        name: Literal[
            "WaPOR3_level_2",
            "WaPOR3_level_3",
            "WaPOR2_level_1",
            "WaPOR2_level_2",
            "WaPOR2_level_3",
            "nrt",
            "all_in",
        ]
        | None = None,
        summary: dict | None = None,
        json: str | None = None,
    ):
        """Load a configuration for the project. Exactly one of `name`, `summary` and
        `json` should be used.

        Parameters
        ----------
        name : Literal["WaPOR3_level_2", "WaPOR3_level_3", "WaPOR2_level_1", "WaPOR2_level_2", "WaPOR2_level_3", "nrt", "all_in"], optional
            Load a predefined configuration, by default None.
        summary : dict, optional
            Load a configuration from a summary, by default None.
        json : str, optional
            Load a configuration from a json-file, by default None.

        Returns
        -------
        Configuration
            Configuration instance.
        """

        if not isinstance(name, type(None)) and not isinstance(summary, type(None)):
            raise ValueError("Only one of `level` and `summary` can be specified.")
        log.info("> CONFIGURATION").add()
        if not isinstance(name, type(None)):
            self.configuration = Configuration.from_name(name)
        elif not isinstance(summary, type(None)):
            summary = summary.copy()
            self.configuration = Configuration.from_summary(summary)
        elif not isinstance(json, type(None)):
            self.configuration = Configuration.from_json(json)
        else:
            raise ValueError(
                "At least one of `level` or `summary` needs to be specified."
            )

        log.info("--> Summary:").add()
        summary_str = self.configuration.__repr__()

        for line in summary_str.split("\n"):
            log.info(line)

        log.sub().sub().info("< CONFIGURATION")
        return self.configuration

    def validate_project_folder(self):
        """Perform several diagnostic tests on the project folder. Mostly to
        detect whether or not different projects have been mixed up in the same
        folder.
        """
        log.info("--> Checking project folder.").add()
        if not os.path.isabs(self.folder):
            self.folder = os.path.abspath(self.folder)
            log.warning(
                f"--> Consider specifying an absolute path to your project folder, continueing with `{self.folder}`."
            )
        if not os.path.isdir(self.folder):
            log.info(
                f"--> Folder does not exist, creating a new folder at `{self.folder}`."
            )
            os.makedirs(self.folder)
        dir_list = os.listdir(self.folder)
        bb = [self.lonlim[0], self.latlim[0], self.lonlim[1], self.latlim[1]]
        check = True

        if "log.txt" in dir_list:
            log_file = os.path.join(self.folder, "log.txt")
            project_folders, periods, bbs, temp_files = Project.parse_log_file(log_file)

            if len(set(project_folders)) > 1:
                log.info("--> Project folder has previously been moved.")
            else:
                ...  # all good.

            if len(set(bbs)) > 1:
                log.warning(
                    "--> You've previously defined different bounding-boxes inside this project folder, this can result in unexpected behaviour."
                )
                check = False
            elif len(set(bbs)) == 1:
                bb_consistent = np.all([np.isclose(x, y) for x, y in zip(bbs[0], bb)])
                if not bb_consistent:
                    log.warning(
                        "--> You've changed the bounding-box of your project without changing the project folder, this can result in unexpected behaviour."
                    )
                    check = False
            else:
                ...  # all good.

            period_ = tuple([np.datetime64(x) for x in self.period])
            if len(set(periods)) > 1:
                log.warning(
                    "--> You've previously defined different periods inside this project folder, this can result in unexpected behaviour."
                )
                check = False
            elif len(set(periods)) == 1:
                period_consistent = period_ == periods[0]
                if not period_consistent:
                    log.warning(
                        "--> You've changed the period of your project without changing the project folder, this can result in unexpected behaviour."
                    )
                    check = False
            else:
                ...  # all good.

            if len(dir_list) > 1:
                self.clean_project_folder()

        elif dir_list:
            log.warning("--> Consider specifying an empty project folder.")
            check = False
        else:
            ...  # all good.

        if check:
            log.info("--> All good.")

        log.sub()

    def clean_project_folder(self):
        """Searched the project folder for corrupt or temporary files that
        couldn't be deleted previously and try again to delete them.
        """

        n_corrupt = n_wrong = n_unkown = n_good = n_temp = 0
        to_remove = list()
        fhs = glob.glob(os.path.join(self.folder, "**/*.nc"), recursive=True)
        log_file = os.path.join(self.folder, "log.txt")

        if os.path.isfile(log_file):
            temp_files = Project.parse_log_file(log_file)[-1]
            for fh in temp_files:
                if fh in fhs:
                    fhs.remove(fh)
                if os.path.isfile(fh):
                    to_remove.append(fh)
                    n_temp += 1

        log.info(f"--> Checking {len(fhs)} files.").add()

        for fh in fhs.copy():
            corrupt = is_corrupt_or_empty(fh)
            if corrupt:
                to_remove.append(fh)
                fhs.remove(fh)
                n_corrupt += 1

        for fh in fhs:
            wrong = has_wrong_bb_or_period(fh, self.bb, self.period)
            if wrong:
                to_remove.append(fh)
                n_wrong += 1
            elif isinstance(wrong, type(None)):
                n_unkown += 1
            else:
                n_good += 1

        if n_temp > 0:
            log.info(f"--> Found {n_temp} temporary files that can be removed.")
        log.info(f"--> Found {n_good} files in good condition.")
        if n_corrupt > 0:
            log.warning(f"--> Found {n_corrupt} corrupt or empty files.")
        if n_wrong > 0:
            log.warning(
                f"--> Found {n_wrong} files with an incorrect `bb` or `period`."
            )
        if n_unkown > 0:
            log.info(f"--> Couldn't determine state of {n_unkown} files.").sub()

        if n_corrupt + n_wrong + n_temp > 0:
            response = input(
                prompt=f"Remove {n_corrupt} corrupt, {n_wrong} wrong and {n_temp} temporary files? (Yes/No)"
            )
            if response in ["Y", "y", "Yes", "yes"]:
                for x in to_remove:
                    remove_ds(x)

    @staticmethod
    def parse_log_file(log_file: str):
        """Subtract some information from the log.txt file.

        Parameters
        ----------
        log_file : str
            Path to log file to parse.

        Returns
        -------
        tuple
            Tuple of defined project folders, periods, bounding-boxes and temporary files
            mentioned in the log file.
        """

        with open(log_file, "r", encoding="utf8") as f:
            log_string = f.read()

        project_folders = re.findall(r"--> Project Folder:\n        > (.*)", log_string)

        periods_ = re.findall(r"--> Period:\n        > (.*) - (.*)", log_string)
        periods = [tuple([np.datetime64(x) for x in period]) for period in periods_]

        bbs_ = re.findall(
            r"--> Bounding-Box:\n\n\s*(\d{1,2}.\d{4})\n.*\n.*\n\s*(\d{1,2}.\d{4})\D*(\d{1,2}.\d{4})\n.*\n.*\n\D*(\d{1,2}.\d{4})",
            log_string,
        )
        bbs = [tuple([[float(x) for x in bb][i] for i in [1, 3, 2, 0]]) for bb in bbs_]

        temp_files = re.findall(
            r"--> Unable to delete temporary file `(.*)`", log_string
        )

        return project_folders, periods, bbs, temp_files

    def set_passwords(self, set_all: bool = False):
        """Set passwords for any datasets that are defined in the configuration.

        Parameters
        ----------
        set_all : bool, optional
            Force setting all possible accounts instead of only the onces
            required for the current configuration, by default `False`.

        """

        log.info("> PASSWORDS").add()

        req_accounts = {
            "MODIS": "NASA",
            "SRTM": "NASA",
            "MERRA2": "NASA",
            "TERRA": "TERRA",
            "ERA5": "CDS",
            "LANDSAT": "EARTHEXPLORER",
            "SENTINEL2": "COPERNICUS_DATA_SPACE",
            "SENTINEL3": "COPERNICUS_DATA_SPACE",
        }

        if not set_all:
            all_accounts = list()
            for v in self.configuration.summary.values():
                if isinstance(v, set):
                    for source_product in v:
                        name = Configuration.source_func(source_product)
                        account = req_accounts.get(name, None)
                        if not isinstance(account, type(None)):
                            all_accounts.append(account)

            x = set(all_accounts)
            log.info(f"--> Accounts needed for `{'`, `'.join(x)}`.").add()
        else:
            x = set(req_accounts.values())
            log.info(f"--> Setting accounts for `{'`, `'.join(x)}`.").add()

        for account in x:
            _ = pywapor.collect.accounts.get(account)

        log.sub().info("--> All set!")
        log.sub().info("< PASSWORDS")

    def download_data(self, buffer_timelim: bool = True):
        """Download data specified in the configuration.

        Parameters
        ----------
        buffer_timelim : bool, optional
            Apply a buffer to the requested period to ensure better interpolation, by default `True`.

        Returns
        -------
        dict
            Overview of collected datasets.
        """
        assert not isinstance(self.configuration, type(None)), (
            "Please load a configuration before continueing."
        )

        log.info("> DOWNLOADER").add()

        example_t_vars = [
            x for x in ["lst", "bt"] if x in self.configuration.full.keys()
        ]
        example_sources = {
            k: v for k, v in self.configuration.full.items() if k in example_t_vars
        }
        other_sources = {
            k: v for k, v in self.configuration.full.items() if k not in example_t_vars
        }

        if buffer_timelim:
            bins = time_bins(self.period, 1)
            adjusted_timelim = [bins[0], bins[-1]]
            buffered_timelim = [
                adjusted_timelim[0] - np.timedelta64(3, "D"),
                adjusted_timelim[1] + np.timedelta64(3, "D"),
            ]
        else:
            adjusted_timelim = self.period
            buffered_timelim = self.period

        example_dss, _ = collect_sources(
            self.folder,
            example_sources,
            self.latlim,
            self.lonlim,
            adjusted_timelim,
            landsat_order_only=True,
        )

        other_dss, _ = collect_sources(
            self.folder,
            other_sources,
            self.latlim,
            self.lonlim,
            buffered_timelim,
        )

        # If there are example-t variables that rely on landsat, try one more time to collect them.
        if np.any(
            list(
                {
                    var: np.any(
                        [
                            product_info["source"] == "LANDSAT"
                            for product_info in info["products"]
                        ]
                    )
                    for var, info in example_sources.items()
                }.values()
            )
        ):
            example_dss, _ = collect_sources(
                self.folder, example_sources, self.latlim, self.lonlim, adjusted_timelim
            )

        self.dss = {**example_dss, **other_dss}

        log.sub().info("< DOWNLOADER")

        return self.dss

    def run_pre_se_root(self, forced: bool = False):
        """Run PRE_SE_ROOT if it hasn't run before.

        Parameters
        ----------
        forced : bool, optional
            Force running PRE_SE_ROOT even if it has already run, by default False.

        Returns
        -------
        xr.Dataset
            Dataset with input for the SE_ROOT model.
        """
        if isinstance(self.se_root_in, type(None)) or forced:
            self.se_root_in = pywapor.pre_se_root.main(
                self.folder,
                self.latlim,
                self.lonlim,
                self.period,
                sources=self.configuration.se_root,
            )
        else:
            log.info("> PRE_SE_ROOT").add()
            log.info(
                f"--> Re-using `{os.path.split(self.se_root_in.encoding['source'])[-1]}`."
            )
            log.sub().info("< PRE_SE_ROOT")
        return self.se_root_in

    def run_se_root(
        self,
        se_root_version="v3",
        export_vars="default",
        chunks={"time": -1, "x": 500, "y": 500},
    ):
        self.se_root_out = pywapor.se_root.main(
            self.se_root_in,
            se_root_version=se_root_version,
            export_vars=export_vars,
            chunks=chunks,
        )
        return self.se_root_out

    def run_pre_et_look(
        self,
        enhancers: List[Callable] = [lapse_rate_to_all],
        bin_length: int | Literal["DEKAD"] = 1,
        forced=False,
    ):
        """Run PRE_ET_LOOK if it hasn't run before.

        Parameters
        ----------
        enhancers : List[Callable], optional
            List of functions to apply to the dataset before returning it. These should generally
            be functions that depend on variables from different sources (otherwise the enhancer
            can be specified in the `configuration` under the relevant variable). For example,
            the lapse rate correction requires temperature variables and elevation data,
            by default [lapse_rate_to_all].
        bin_length : int | Literal["DEKAD"], optional
            Number of days over which to aggregate the output data, should at least
            be 1. If specified as `"DEKAD"`, the data will be aggregated over three dekads
            per month, by default 1.
        forced : bool, optional
            Force running PRE_SE_ROOT even if it has already run, by default False.

        Returns
        -------
        xr.Dataset
            Dataset with input for the ET_LOOK model.
        """
        if isinstance(self.et_look_in, type(None)) or forced:
            self.et_look_in = pywapor.pre_et_look.main(
                self.folder,
                self.latlim,
                self.lonlim,
                self.period,
                sources=self.configuration.et_look,
                enhancers=enhancers,
                bin_length=bin_length,
            )
        else:
            log.info("> PRE_ET_LOOK").add()
            log.info(
                f"--> Re-using `{os.path.split(self.et_look_in.encoding['source'])[-1]}`."
            )
            log.sub().info("< PRE_ET_LOOK")
        return self.et_look_in

    def run_et_look(
        self,
        et_look_version: Literal["v2", "v3"] = "v3",
        export_vars: Literal["default", "all"] | List[str] = "default",
        chunks: Dict[str, int] = {"time_bins": -1, "x": 500, "y": 500},
    ):
        """Run the ET_LOOK model.

        Parameters
        ----------
        et_look_version : Literal["v2", "v3"], optional
            Choose which version of the model to run, by default "v3".
        export_vars : Literal["default", "all"] | List[str], optional
            Choose which variables to export, `"default"` exports the most important variables,
            `"all"` exports all variables that have been calculated. Pass a list with variable names
            to control precisely which variables get written into the output file, by default "default".
        chunks : Dict[str, int], optional
            Choose how the data is chunked during calculation. Check
            `https://docs.xarray.dev/en/stable/user-guide/dask.html` for more
            information, by default {"time_bins": -1, "x": 500, "y": 500}.

        Returns
        -------
        xr.Dataset
            Dataset with output of the ET_LOOK model.
        """
        self.et_look_out = pywapor.et_look.main(
            self.et_look_in,
            et_look_version=et_look_version,
            export_vars=export_vars,
            chunks=chunks,
        )
        return self.et_look_out


if __name__ == "__main__":
    timelim = ["2023-11-01", "2023-11-30"]
    latlim = [21.9692194682626933, 21.9939120838340507]
    lonlim = [91.9371349243682801, 91.9657566608824339]
    bb = [
        91.9371349243682801,
        21.9692194682626933,
        91.9657566608824339,
        21.9939120838340507,
    ]
    project_folder = r"/Users/hmcoerver/Local/pywapor_bgdX"

    adjust_logger(True, project_folder, "INFO")

    all_configs = {
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR3_level_2",
        "WaPOR3_level_3",
        "nrt",
        "all_in",
    }

    project = pywapor.Project(project_folder, bb, timelim)

    # project.validate_project_folder()

    project.load_configuration(name="WaPOR3_level_2")

    summary = {
        # Define which products to use.
        "elevation": {},
        "meteorological": {},
        "optical": {"LANDSAT.LC09_SR"},
        "precipitation": {},
        "solar radiation": {},
        "statics": {},
        "thermal": {},
        "soil moisture": {},
        # Define which product to reproject the other products to.
        "_EXAMPLE_": "LANDSAT.LC09_SR",
        # # Define any special functions to apply to a specific variable.
        # '_ENHANCE_': {"bt": ["pywapor.enhancers.dms.thermal_sharpener.sharpen"],},
        # # Choose which products should be gapfilled.
        # '_WHITTAKER_': {
        #     'SENTINEL2.S2MSI2A_R20m': {'lmbdas': 1000.0, 'method': 'whittaker'},
        #     'VIIRSL1.VNP02IMG': {'a': 0.85, 'lmbdas': 1000.0, 'method': 'whittaker'}},
    }

    # project.load_configuration(summary = summary)

    # project.set_passwords()
    # dss = project.download_data()

    # se_root_in = project.run_pre_se_root()
    # se_root = project.run_se_root()
    # et_look_in = project.run_pre_et_look()
    # et_look = project.run_et_look()



import copy
import datetime
import os

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup
from osgeo import gdal

import pywapor.collect.protocol.cog as cog
from pywapor.collect.protocol.projections import get_crss
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.logger import log
from pywapor.general.processing_functions import open_ds, process_ds, remove_ds, save_ds


def default_vars(product_name, req_vars=["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "P05": {
            "precip": [("time", "lat", "lon"), "p"],
            "crs": [(), "spatial_ref"],
        }
    }

    req_dl_vars = {
        "P05": {
            "p": ["precip", "crs"],
        }
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def default_post_processors(product_name, req_vars=["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "P05": {
            "p": [],
        }
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def most_recent(product_name, *args):
    today = datetime.date.today()
    year = today.year + 1

    nothing_found = True
    while nothing_found and year > 1900:
        year = year - 1
        url = f"https://data.chc.ucsb.edu/products/CHIRPS-2.0/global_daily/cogs/{product_name.lower()}/{year}"
        r = requests.get(url)
        soup = BeautifulSoup(r.content, features="xml")
        x = soup.findAll(lambda tag: tag.name == "a" and ".cog" in tag.text)
        if len(x) > 0:
            nothing_found = False
    recent = datetime.datetime.strptime(x[-1].get_text(), "chirps-v2.0.%Y.%m.%d.cog")

    return recent


def download(
    folder,
    latlim,
    lonlim,
    timelim,
    product_name="P05",
    req_vars=["p"],
    variables=None,
    post_processors=None,
):
    """Download CHIRPS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "P05".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["p"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "CHIRPS")

    if not os.path.isdir(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.05, latlim[1] + 0.05]
        lonlim = [lonlim[0] - 0.05, lonlim[1] + 0.05]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    timedelta = pd.Timedelta(hours=12)
    data_source_crs = get_crss("WGS84")

    chirps_domain = [
        -180.0000000000000000,
        -50.0000014901161194,
        180.0000053644180298,
        50.0000000000000000,
    ]
    chirps_res = [7200, 2000]
    chirps_px_size = [
        (chirps_domain[2] - chirps_domain[0]) / chirps_res[0],
        (chirps_domain[3] - chirps_domain[1]) / chirps_res[1],
    ]

    def _snap_point(x0, xres, X):
        n = np.round((X - x0) / xres)
        Xsnap = n * xres + x0
        return Xsnap, n

    latlim_ = [
        _snap_point(chirps_domain[0], chirps_px_size[0], latlim[0])[0],
        _snap_point(chirps_domain[0], chirps_px_size[0], latlim[1])[0],
    ]
    lonlim_ = [
        _snap_point(chirps_domain[1], chirps_px_size[1], lonlim[0])[0],
        _snap_point(chirps_domain[1], chirps_px_size[1], lonlim[1])[0],
    ]

    warp_kwargs = dict()
    warp_kwargs["outputBounds"] = [lonlim_[0], latlim_[0], lonlim_[1], latlim_[1]]
    warp_kwargs["outputBoundsSRS"] = "epsg:4326"

    dates = pd.date_range(timelim[0], timelim[1], freq="D")
    urls = [
        f"/vsicurl/https://data.chc.ucsb.edu/products/CHIRPS-2.0/global_daily/cogs/p05/{x.year}/chirps-v2.0.{x.year}.{x.month:>02}.{x.day:>02}.cog"
        for x in dates
    ]
    temp_fn = fn.replace(".nc", "_temp.nc")
    temp_vrt = temp_fn.replace(".nc", ".vrt")

    gdal_config_options = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
    }
    try:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, v)
        x = cog.cog_dl(urls, temp_fn, warp_kwargs=warp_kwargs)
    except Exception as e:
        raise e
    finally:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, None)

    ds_ = open_ds(x)
    ds = ds_.to_array("time").to_dataset(name="precip")

    if ds["time"].size != len(urls):
        invalid_urls = cog.ping_urls(urls)
        idxs = [urls.index(invalid_url) for invalid_url in invalid_urls]
        new_dates = [x for i, x in enumerate(dates) if i not in idxs]
        log.warning(
            f"The following dates could not be collected: {set(new_dates).symmetric_difference(set(dates))}"
        )
        dates = new_dates

    ds = ds.assign_coords({"time": [np.datetime64(x + timedelta, "ns") for x in dates]})
    ds = process_ds(ds, coords, variables, crs=data_source_crs)

    ds = apply_enhancers(post_processors, ds)

    ds = save_ds(ds, fn, label="Merging files.")

    remove_ds(temp_fn)
    if os.path.isfile(temp_vrt):
        remove_ds(temp_vrt)

    return ds[req_vars_orig]



"""
https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NUMNC.003/contents.html
"""

import copy
import datetime
import json
import os

from shapely.geometry import shape
from shapely.geometry.polygon import Polygon

import pywapor.collect
import pywapor.collect.accounts as accounts
import pywapor.collect.protocol.opendap as opendap
from pywapor.enhancers.dem import calc_slope_or_aspect
from pywapor.general.processing_functions import open_ds


def tiles_intersect(latlim, lonlim):
    """Creates a list of server-side filenames for tiles that intersect with `latlim` and
    `lonlim` for the selected product.

    Parameters
    ----------
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.

    Returns
    -------
    list
        Server-side filenames for tiles.
    """
    with open(
        os.path.join(pywapor.collect.__path__[0], "product/SRTM30_tiles.geojson")
    ) as f:
        features = json.load(f)["features"]
    aoi = Polygon.from_bounds(lonlim[0], latlim[0], lonlim[1], latlim[1])
    tiles = list()
    for feature in features:
        shp = shape(feature["geometry"])
        tile = feature["properties"]["dataFile"]
        if shp.intersects(aoi):
            tiles.append(tile.split(".")[0])
    return tiles


def default_vars(product_name, req_vars=["z"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "30M": {
            "SRTMGL1_DEM": [("time", "lat", "lon"), "z"],
            "crs": [(), "spatial_ref"],
        }
    }

    req_dl_vars = {
        "30M": {
            "z": ["SRTMGL1_DEM", "crs"],
            "slope": ["SRTMGL1_DEM", "crs"],
            "aspect": ["SRTMGL1_DEM", "crs"],
        }
    }

    out = {
        val: variables[product_name][val]
        for sublist in map(req_dl_vars[product_name].get, req_vars)
        for val in sublist
    }

    return out


def drop_time(ds):
    return ds.isel(time=0).drop("time")


def default_post_processors(product_name, req_vars=["z"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "30M": {
            "z": [],
            "aspect": [calc_slope_or_aspect],
            "slope": [calc_slope_or_aspect],
        }
    }

    out = {k: v for k, v in post_processors[product_name].items() if k in req_vars}

    return out


def fn_func(product_name, tile):
    """Returns a client-side filename at which to store data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_{tile}.nc"
    return fn


def url_func(product_name, tile):
    """Returns a url at which to collect MERRA2 data.

    Parameters
    ----------
    product_name : None
        Not used.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        The url.
    """
    url = f"https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NC.003/{tile}.SRTMGL1_NC.ncml.nc4?"
    return url

def most_recent(product_name, *args):
    return None

def download(
    folder,
    latlim,
    lonlim,
    product_name="30M",
    req_vars=["z"],
    variables=None,
    post_processors=None,
    **kwargs,
):
    """Download SRTM data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "30M".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["z"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "SRTM")

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.0002777777777777768
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    timelim = [datetime.date(2000, 2, 10), datetime.date(2000, 2, 12)]
    tiles = tiles_intersect(latlim, lonlim)

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {
            k: {True: default_processors[k], False: v}[v == "default"]
            for k, v in post_processors.items()
            if k in req_vars
        }

    data_source_crs = None
    parallel = False
    spatial_tiles = True
    un_pw = accounts.get("NASA")
    request_dims = True

    ds = opendap.download(
        fn,
        product_name,
        coords,
        variables,
        post_processors,
        fn_func,
        url_func,
        un_pw=un_pw,
        tiles=tiles,
        data_source_crs=data_source_crs,
        parallel=parallel,
        spatial_tiles=spatial_tiles,
        request_dims=request_dims,
    )

    return ds[req_vars_orig]


if __name__ == "__main__":
    folder = r"/Users/hmcoerver/Local/cog2_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]
    req_vars = ["z", "aspect", "slope"]

    variables = None
    post_processors = None

    # fn = os.path.join(os.path.join(folder, "SRTM"), "30M.nc")
    # if os.path.isfile(fn):
    #     os.remove(fn)

    # SRTM.
    ds = download(folder, latlim, lonlim, req_vars=req_vars)
    print(ds.rio.crs, ds.rio.grid_mapping)



# -*- coding: utf-8 -*-
"""
Code to run the ETLook model.
"""
import os
import pywapor.et_look_v2_v3 as ETLook_v2_v3
import pywapor.general as g
import datetime
from pywapor.general.logger import log, adjust_logger
from pywapor.general.processing_functions import save_ds, open_ds
import copy
import warnings

def main(input_data, et_look_version = "v2", export_vars = "default", chunks = {"time_bins": -1, "x": 500, "y": 500}):
    """Runs the ETLook model over the provided input data.

    Parameters
    ----------
    input_data : str | xr.Dataset
        Dataset generated by `pywapor.pre_et_look`.
    et_look_version : "v2" | "v3", optional
        Which version of the ETLook model to use, by default "v2".
    export_vars : "default" | "all" | list, optional
        Specify which variables to save inside the output file. `"Default"` stores `int_mm`,
        `t_24_mm`, `e_24_mm`, `aeti_24_mm`, `et_ref_24_mm`, `se_root`, `biomass_prod`,
        `epoch_ends` and `epoch_starts`. `"all"` stores all calculated variables. Use a
        list to specify a custom output set, by default "default".
    chunks : dict, optional
        Specify how the calculations are split up. Increase chunk sizes to speed up calculation, 
        decrease to use less RAM, by default {"time": 1, "x": 1000, "y": 1000}.


    Returns
    -------
    xr.Dataset
        Dataset with variables selected through `export_vars`.
    """

    # Inputs
    if isinstance(input_data, str):
        ds = open_ds(input_data, chunks = chunks)
    else:
        ds = copy.deepcopy(input_data).chunk(chunks)
        input_data = ds.encoding["source"]

    _ = adjust_logger(True, os.path.split(input_data)[0], "INFO")

    t1 = datetime.datetime.now()
    log.info("> ET_LOOK").add()

    # Version
    ETLook = ETLook_v2_v3

    log.info(f"--> Running `et_look` ({et_look_version}).")

    # Allow skipping of et_look-functions if not all of its required inputs are
    # available.
    g.lazifier.decorate_submods(ETLook, g.lazifier.etlook_decorator)

    ds = g.variables.initiate_ds(ds)

    if ds["rs_min"].dtype == object:
        ds["rs_min"] = 100
        log.info("--> Setting `rs_min` to `100`.")

    if ds["land_mask"].dtype == object:
        ds["land_mask"] = 1
        log.info("--> Setting `land_mask` to `1`.")

    if ds["z_obst_max"].dtype == object:
        ds["z_obst_max"] = 3
        log.info("--> Setting `z_obst_max` to `3`.")

    ds["decl"] = ETLook.solar_radiation.declination(ds["doy"])
    ds["iesd"] = ETLook.solar_radiation.inverse_earth_sun_distance(ds["doy"])

    ######################## MODEL ETLOOK ####################################

    # **effective_leaf_area_index*********************************************
    ds["vc"] = ETLook.leaf.vegetation_cover(ds["ndvi"], nd_min = ds["nd_min"], nd_max = ds["nd_max"], vc_pow = ds["vc_pow"])
    ds["lai"] = ETLook.leaf.leaf_area_index(ds["vc"], vc_min = ds["vc_min"], vc_max = ds["vc_max"], lai_pow = ds["lai_pow"])
    ds["lai_eff"] = ETLook.leaf.effective_leaf_area_index(ds["lai"])
    
    # *******TRANSPIRATION COMPONENT******************************************
    
    # **soil fraction*********************************************************
    ds["sf_soil"] = ETLook.radiation.soil_fraction(ds["lai"])

    # **atmospheric canopy resistance***********************************************
    ds["lat_rad"] = ETLook.solar_radiation.latitude_rad(ds["y"]).chunk("auto")
    ds["ws"] = ETLook.solar_radiation.sunset_hour_angle(ds["lat_rad"], ds["decl"])

    ds["ra_toa_flat_24"] = ETLook.solar_radiation.daily_solar_radiation_toa_flat(ds["decl"], ds["iesd"], ds["lat_rad"], ds["ws"])

    if ds["slope"].dtype == object or ds["aspect"].dtype == object:
        ds["ra_24"] = ds["ra_flat_24"]
        ds["trans_24"] = ETLook.solar_radiation.transmissivity(ds["ra_flat_24"], ds["ra_toa_flat_24"])
    else:
        ds["sc"] = ETLook.solar_radiation.seasonal_correction(ds["doy"])
        ds["ra_toa_24"] = ETLook.solar_radiation.daily_solar_radiation_toa(ds["sc"], ds["decl"], ds["iesd"], ds["lat_rad"], ds["slope"], ds["aspect"])
        ds["trans_24"] = ETLook.solar_radiation.transmissivity(ds["ra_flat_24"], ds["ra_toa_flat_24"])
        ds["diffusion_index"] = ETLook.solar_radiation.diffusion_index(ds["trans_24"], diffusion_slope = ds["diffusion_slope"], diffusion_intercept = ds["diffusion_intercept"])
        ds["ra_24"] = ETLook.solar_radiation.daily_total_solar_radiation(ds["ra_toa_24"], ds["ra_toa_flat_24"], ds["diffusion_index"], ds["trans_24"])

    ds["stress_rad"] = ETLook.stress.stress_radiation(ds["ra_24"])
    ds["p_air_0_24_mbar"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_0_24"])
    ds["p_air_24"] = ETLook.meteo.air_pressure_daily(ds["z"], ds["p_air_0_24_mbar"])

    if ds["vp_24"].dtype == object:
        ds["vp_24"] = ETLook.meteo.vapour_pressure_from_specific_humidity_daily(ds["qv_24"], ds["p_air_24"])
    if ds["vp_24"].dtype == object:
        ds["vp_24"] = ETLook.meteo.vapour_pressure_from_dewpoint_daily(ds["t_dew_24"])

    if et_look_version == "v2":
        ds["svp_24"] = ETLook.meteo.saturated_vapour_pressure(ds["t_air_24"])
    else:
        ds["svp_24_min"] = ETLook.meteo.saturated_vapour_pressure_minimum(ds["t_air_min_24"])
        ds["svp_24_max"] = ETLook.meteo.saturated_vapour_pressure_maximum(ds["t_air_max_24"])
        ds["svp_24"] = ETLook.meteo.saturated_vapour_pressure_average(ds["svp_24_max"], ds["svp_24_min"])

    ds["vpd_24"] = ETLook.meteo.vapour_pressure_deficit_daily(ds["svp_24"], ds["vp_24"])
    ds["stress_vpd"] = ETLook.stress.stress_vpd(ds["vpd_24"], ds["vpd_slope"])
    ds["stress_temp"] = ETLook.stress.stress_temperature(ds["t_air_24"], t_opt = ds["t_opt"], t_min = ds["t_min"], t_max = ds["t_max"])

    ds["r_canopy_0"] = ETLook.resistance.atmospheric_canopy_resistance(ds["lai_eff"], ds["stress_rad"], ds["stress_vpd"], ds["stress_temp"], rs_min = ds["rs_min"], rcan_max = ds["rcan_max"])

    # **net radiation canopy******************************************************
    ds["t_air_k_24"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_24"])
    ds["l_net"] = ETLook.radiation.longwave_radiation_fao(ds["t_air_k_24"], ds["vp_24"], ds["trans_24"], vp_slope = ds["vp_slope"], vp_offset = ds["vp_offset"], lw_slope = ds["lw_slope"], lw_offset = ds["lw_offset"])
    ds["int_mm"] = ETLook.evapotranspiration.interception_mm(ds["p_24"], ds["vc"], ds["lai"], int_max = ds["int_max"])
    ds["lh_24"] = ETLook.meteo.latent_heat_daily(ds["t_air_24"])
    ds["int_wm2"] = ETLook.radiation.interception_wm2(ds["int_mm"], ds["lh_24"])
    ds["rn_24"] = ETLook.radiation.net_radiation(ds["r0"], ds["ra_24"], ds["l_net"], ds["int_wm2"])
    ds["rn_24_canopy"] = ETLook.radiation.net_radiation_canopy(ds["rn_24"], ds["sf_soil"])

    ds["stress_moist"] = ETLook.stress.stress_moisture(ds["se_root"], tenacity = ds["tenacity"])
    ds["r_canopy"] = ETLook.resistance.canopy_resistance(ds["r_canopy_0"], ds["stress_moist"], rcan_max = ds["rcan_max"])

    # **initial canopy aerodynamic resistance***********************************************************
    if ds["z_oro"].dtype == object: # TODO this function is wrong...
        ds["z_oro"] = 0.001
        log.info("--> Setting `z_oro` to `0.001`.")
        # ds["z_oro"] = ETLook.roughness.orographic_roughness(ds["z"], ds["x"], ds["y"])
    
    ds["z_obst"] = ETLook.roughness.obstacle_height(ds["ndvi"], ds["z_obst_max"], ndvi_obs_min = ds["ndvi_obs_min"], ndvi_obs_max = ds["ndvi_obs_max"], obs_fr = ds["obs_fr"])
    ds["z0m"] = ETLook.roughness.roughness_length(ds["lai"], ds["z_oro"], ds["z_obst"], ds["z_obst_max"], ds["land_mask"])
    if ds["u_24"].dtype == object:
        ds["u_24"] = ETLook.meteo.wind_speed(ds["u2m_24"], ds["v2m_24"])
    ds["ra_canopy_init"] = ETLook.neutral.initial_canopy_aerodynamic_resistance(ds["u_24"], ds["z0m"], z_obs = ds["z_obs"])

    # **windspeed blending height daily***********************************************************
    ds["u_b_24"] = ETLook.meteo.wind_speed_blending_height_daily(ds["u_24"], z_obs = ds["z_obs"], z_b = ds["z_b"])

    # **ETLook.neutral.initial_daily_transpiration***********************************************************
    ds["ad_dry_24"] = ETLook.meteo.dry_air_density_daily(ds["p_air_24"], ds["vp_24"], ds["t_air_k_24"])
    ds["ad_moist_24"] = ETLook.meteo.moist_air_density_daily(ds["vp_24"], ds["t_air_k_24"])
    ds["ad_24"] = ETLook.meteo.air_density_daily(ds["ad_dry_24"], ds["ad_moist_24"])
    ds["psy_24"] = ETLook.meteo.psychrometric_constant_daily(ds["p_air_24"], ds["lh_24"])
    ds["ssvp_24"] = ETLook.meteo.slope_saturated_vapour_pressure_daily(ds["t_air_24"])
    ds["t_24_init"] = ETLook.neutral.initial_daily_transpiration(ds["rn_24_canopy"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_canopy"], ds["ra_canopy_init"])

    # **ETLook.unstable.initial_sensible_heat_flux_canopy_daily***********************************************************
    ds["h_canopy_24_init"] = ETLook.unstable.initial_sensible_heat_flux_canopy_daily(ds["rn_24_canopy"], ds["t_24_init"])

    # **ETLook.unstable.initial_friction_velocity_daily***********************************************************
    ds["disp"] = ETLook.roughness.displacement_height(ds["lai"], ds["z_obst"], land_mask = ds["land_mask"], c1 = ds["c1"])
    ds["u_star_24_init"] = ETLook.unstable.initial_friction_velocity_daily(ds["u_b_24"], ds["z0m"], ds["disp"], z_b = ds["z_b"])

    # **ETLook.unstable.transpiration***********************************************************
    ds["t_24"] = ETLook.unstable.transpiration(ds["rn_24_canopy"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_canopy"], ds["h_canopy_24_init"], ds["t_air_k_24"], ds["u_star_24_init"], ds["z0m"], ds["disp"], ds["u_b_24"], z_obs = ds["z_obs"], z_b = ds["z_b"], iter_h = ds["iter_h"])
    ds["t_24_mm"] = ETLook.unstable.transpiration_mm(ds["t_24"], ds["lh_24"])

    #*******EVAPORATION COMPONENT****************************************************************

    # **ETLook.radiation.net_radiation_soil***********************************************************
    ds["rn_24_soil"] = ETLook.radiation.net_radiation_soil(ds["rn_24"], ds["sf_soil"])

    # **ETLook.resistance.soil_resistance***********************************************************
    ds["r_soil"] = ETLook.resistance.soil_resistance(ds["se_root"], land_mask = ds["land_mask"], r_soil_pow = ds["r_soil_pow"], r_soil_min = ds["r_soil_min"])

    # **ETLook.resistance.soil_resistance***********************************************************
    ds["ra_soil_init"] = ETLook.neutral.initial_soil_aerodynamic_resistance(ds["u_24"], z_obs = ds["z_obs"])

    # **ETLook.unstable.initial_friction_velocity_soil_daily***********************************************************
    ds["u_star_24_soil_init"] = ETLook.unstable.initial_friction_velocity_soil_daily(ds["u_b_24"], ds["disp"], z_b = ds["z_b"])

    # **ETLook.unstable.initial_sensible_heat_flux_soil_daily***********************************************************
    ds["stc"] = ETLook.radiation.soil_thermal_conductivity(ds["se_root"])
    ds["vhc"] = ETLook.radiation.volumetric_heat_capacity(ds["se_root"], porosity = ds["porosity"])    
    
    ds["dd"] = ETLook.radiation.damping_depth(ds["stc"], ds["vhc"])
    ds["g0_bs"] = ETLook.radiation.bare_soil_heat_flux(ds["doy"], ds["dd"], ds["stc"], ds["t_amp"], ds["lat_rad"])
    ds["g0_24"] = ETLook.radiation.soil_heat_flux(ds["g0_bs"], ds["sf_soil"], ds["land_mask"], ds["rn_24_soil"], ds["trans_24"], ds["ra_24"], ds["l_net"], ds["rn_slope"], ds["rn_offset"])
    ds["e_24_init"] = ETLook.neutral.initial_daily_evaporation(ds["rn_24_soil"], ds["g0_24"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_soil"], ds["ra_soil_init"])
    ds["h_soil_24_init"] = ETLook.unstable.initial_sensible_heat_flux_soil_daily(ds["rn_24_soil"], ds["e_24_init"], ds["g0_24"])

    # **ETLook.unstable.evaporation***********************************************************
    ds["e_24"] = ETLook.unstable.evaporation(ds["rn_24_soil"], ds["g0_24"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_soil"], ds["h_soil_24_init"], ds["t_air_k_24"], ds["u_star_24_soil_init"], ds["disp"], ds["u_b_24"], z_b = ds["z_b"], z_obs = ds["z_obs"], iter_h = ds["iter_h"])
    ds["e_24_mm"] = ETLook.unstable.evaporation_mm(ds["e_24"], ds["lh_24"])
    ds["aeti_24_mm"] = ETLook.evapotranspiration.eti_actual_mm(ds["e_24_mm"], ds["t_24_mm"], ds["int_mm"])

    # **ETLook.unstable.evaporation***********************************************************
    ds["rn_24_grass"] = ETLook.radiation.net_radiation_grass(ds["ra_24"], ds["l_net"], r0_grass = ds["r0_grass"])
    ds["et_ref_24"] = ETLook.evapotranspiration.et_reference(ds["rn_24_grass"], ds["ad_24"], ds["psy_24"], ds["vpd_24"], ds["ssvp_24"], ds["u_24"])
    ds["et_ref_24_mm"] = ETLook.evapotranspiration.et_reference_mm(ds["et_ref_24"], ds["lh_24"])

    ds["t_air_k_min"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_min_24"])
    ds["t_air_k_max"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_max_24"])

    ds["t_air_k_12"] = ETLook.meteo.mean_temperature_kelvin_daytime(ds["t_air_k_min"], ds["t_air_k_max"])

    ds["t_dep"] = ETLook.biomass.temperature_dependency(ds["t_air_k_12"], dh_ap=ds["dh_ap"], d_s=ds["d_s"], dh_dp=ds["dh_dp"])
    ds["k_m"] = ETLook.biomass.affinity_constant_co2(ds["t_air_k_12"])
    ds["k_0"] = ETLook.biomass.inhibition_constant_o2(ds["t_air_k_12"])
    ds["tau_co2_o2"] = ETLook.biomass.co2_o2_specificity_ratio(ds["t_air_k_12"])

    ds["year"] = ds.time_bins.dt.year.chunk("auto")

    ds["co2_act"] = ETLook.biomass.co2_level_annual(ds["year"])
    ds["a_d"] = ETLook.biomass.autotrophic_respiration(ds["t_air_k_24"], ar_slo=ds["ar_slo"], ar_int=ds["ar_int"])

    ds["apar"] = ETLook.biomass.par(ds["ra_24"])
    ds["f_par"] = ETLook.biomass.fpar(ds["ndvi"], fpar_slope=ds["fpar_slope"], fpar_offset=ds["fpar_offset"])

    ds["co2_fert"] = ETLook.biomass.co2_fertilisation(ds["tau_co2_o2"], ds["k_m"], ds["k_0"], ds["co2_act"], o2=ds["o2"], co2_ref=ds["co2_ref"])
    ds["npp_max"] = ETLook.biomass.net_primary_production_max(ds["t_dep"], ds["co2_fert"], ds["a_d"], ds["apar"], gcgdm=ds["gcgdm"])
    ds["npp"] = ETLook.biomass.net_primary_production(ds["npp_max"], ds["f_par"], ds["stress_moist"], phot_eff=ds["phot_eff"]) # TODO include phot_eff from table (based on LULC)

    ds = ds.drop_vars([x for x in ds.variables if ds[x].dtype == object])

    fp, fn = os.path.split(input_data)

    ds = g.variables.fill_attrs(ds)

    if export_vars == "all":
        ...
    elif export_vars == "default":
        keep_vars = [
                    'int_mm',
                    't_24_mm',
                    'e_24_mm',
                    'aeti_24_mm',
                    'se_root',
                    'npp'
                    ]
        keep_vars = [x for x in keep_vars if x in ds.variables]
        ds = ds[keep_vars]
    elif isinstance(export_vars, list):
        keep_vars = copy.copy(export_vars)
        ds = ds[keep_vars]
    else:
        raise ValueError("Please provide a valid `export_vars` ('all', 'default' or a list).")

    if len(ds.data_vars) == 0:
        log.info("--> No data to export, try adjusting `export_vars` or make sure to provide required inputs.")
        ds = None
    else:
        ds = ds.transpose("time_bins", "y", "x") # set dimension order the same for all vars.
        fn = fn.replace("in", "out")
        fp_out = os.path.join(fp, fn)
        if os.path.isfile(fp_out):
            fp_out = fp_out.replace(".nc", "_.nc")
        
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="invalid value encountered in power")
            warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
            warnings.filterwarnings("ignore", message="invalid value encountered in divide")
            warnings.filterwarnings("ignore", message="divide by zero encountered in power")
            warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
            warnings.filterwarnings("ignore", message="divide by zero encountered in log")
            warnings.filterwarnings("ignore", message="divide by zero encountered in divide")
            ds = save_ds(ds, fp_out, encoding = "initiate", chunks = chunks, label = f"Saving output to `{fn}`.")

    t2 = datetime.datetime.now()
    log.sub().info(f"< ET_LOOK ({str(t2 - t1)})")

    return ds

def check_for_non_chuncked_arrays(ds):
    for var in ds.data_vars:
        if len(ds[var].dims) > 0:
            if isinstance(ds[var].chunks, type(None)):
                print(var)

if __name__ == "__main__":
    ...
    chunks = {"time_bins": -1, "x": 1000, "y": 1000}
    et_look_version = "v3"
    export_vars = "default"

    # input_data = '/Users/hmcoerver/Local/pywapor_se/et_look_in.nc'

    # input_data = r'/Users/hmcoerver/Local/test8/et_look_in_.nc'
    # et_look_version = "v2"
    # export_vars = "default"

    # out = main(input_data, et_look_version = et_look_version, export_vars = export_vars)